.setcpu "65816"
.smart

;Included Files

.include "Header.asm"

;Define Stuff

HVBJOY=		$4212
JOY0=		$4218
JOY1=		$421a
JOY2=		$421c
JOY3=		$421e
JOYSER0=	$4016
JOYSER1=	$4017
WRIO=		$4201

;Button Numbers
RButton=	16
LButton= 	32
XButton= 	64
AButton= 	128

RightDPad=	1
LeftDPad=	2
DownDPad=	4
UpDPad=		8
StartButton=	16
SelectButton=	32
YButton=	64
BButton=	128

WordBitRightDPad=	256
WordLeftDPad=		512
WordDownDPad=		1024
WordUpDPad=		2048
WordStartButton=	4096
WordSelectButton=	8192
WordYButton=		16384
WordBButton=		32768

;Game variables, located within 6502 zero page ($00-ff)
;65816 has direct page which can range from $0000-ffff in region, but ca65
;only seems to understand "zp" being up to 256 bytes in size.
;See lorom256k.cfg and lda/tcd code in _reset

;=========================================================================================
.segment "ZEROPAGE"
;=========================================================================================

Joy1Data:		.res 2
Joy2Data:		.res 2
Joy1Press:		.res 2
Joy2Press:		.res 2
SpriteCount:		.res 2

GameState:		.res 2
BallXPosition:		.res 2
BallYPosition:		.res 2
BallXVelocity:		.res 2
BallYVelocity:		.res 2
Paddle1YPosition:	.res 2
Paddle2YPosition:	.res 2

;SNES RAM area, $0200-1fff
;See snes.cfg

;=========================================================================================
.segment "BSS"
;=========================================================================================

SpriteBuf1:		.res $200	;Sprite buffer for OAM DMA, 512 ($200) bytes
SpriteBuf1Size=		*-SpriteBuf1
SpriteBuf2:    		.res $20	;Sprite buffer for OAM DMA, 32  ($20) bytes
SpriteBuf2Size= 	*-SpriteBuf2
SpriteBuf3:		.res $200	;Sprite buffer for OAM DMA, 512  ($200) bytes
SpriteBuf3Size=		*-SpriteBuf3
SpriteBufSize=		SpriteBuf1Size+SpriteBuf2Size

;=========================================================================================
.segment "BSS7E"
;=========================================================================================

;Assemble-time assertion; SNES DMA transfer can only handle 65536 bytes.
.assert (SpriteBufSize < $10000), error, "SpriteBufs area too big for SNES DMA transfer"

;=========================================================================================
.segment "CODE"
;=========================================================================================

.proc Main

  sei		;Disable interrupts
  clc		;Clear carry, used by XCE
  xce		;Set 65816 native mode
  jmp :+	;Needed to set K (bank of PC) properly if MODE 21 is ever used;
		;see official SNES developers docs, "Programming Cautions"

: cld		;Disable decimal mode
  phk		;Push K (PC bank)
  plb		;Pull B (data bank, i.e. data bank now equals PC bank)

  rep #$30	;A=16, X/Y=16

;Note: this should correlate with ZEROPAGE in lorom256k.cfg
  lda #$0000
  tcd		;Set D = $0000 (direct page)

;Note: this should correlate with the top of BSS in lorom256k.cfg
  ldx #$1fff
  txs		;Set X = $1fff (stack pointer)

;Register initialisation values, per official Nintendo documentation
;Note: if you are looking at this code and thinking "how can I optimise"
;this?", then you are doing nothing useful with your time.  This code is
;called ONCE on reset or power-on, it does not need to be optimised.  I
;(koitsu) find it remarkable how many people have tried to "optimise"
;this routine.  STOP IT.

  sep #$20	;A=8
  lda #$80
  sta $2100
  stz $2101
  stz $2102
  stz $2103
  stz $2104
  stz $2105
  stz $2106
  stz $2107
  stz $2108
  stz $2109
  stz $210A
  stz $210B
  stz $210C
  stz $210D
  stz $210D
  stz $210E
  stz $210E
  stz $210F
  stz $210F
  stz $2110
  stz $2110
  stz $2111
  stz $2111
  stz $2112
  stz $2112
  stz $2113               
  stz $2113               
  stz $2114
  stz $2114
  lda #$80
  sta $2115
  stz $2116
  stz $2117
  stz $211A
  stz $211B
  lda #$01
  sta $211B
  stz $211C
  stz $211C
  stz $211D
  stz $211D
  stz $211E       
  lda #$01
  sta $211E
  stz $211F
  stz $211F
  stz $2120
  stz $2120
  stz $2121
  stz $2123
  stz $2124
  stz $2125
  stz $2126
  stz $2127
  stz $2128
  stz $2129
  stz $212A
  stz $212B
  stz $212C
  stz $212D
  stz $212E
  stz $212F
  stz $4200
  lda #$FF
  sta $4201
  stz $4202
  stz $4203
  stz $4204
  stz $4205
  stz $4206
  stz $4207
  stz $4208
  stz $4209
  stz $420A
  stz $420B
  stz $420C
  stz $420D

  jsr clear_vram	;Zero VRAM. Requires A=8, X/Y=16
  jsr clear_cgram	;Zero palette. Requires A=8, X/Y=16
  
;clear sprite tables

  stz $2102	;sprites initialized to be off the screen, palette 0, character 0
  stz $2103
  ldx #$0080
  lda #$F0

loop08:
  sta $2104	;set X = 240
  sta $2104	;set Y = 240
  sta $2104	;set character = $00
  sta $2104	;set priority=0, no flips
  dex
  bne loop08
  ldx #$0020

loop09:
  stz $2104		;set size bit=0, x MSB = 0
  dex
  bne loop09

;clear SNES RAM

  stz $2181	;set WRAM address to $000000
  stz $2182
  stz $2183

  ldx #$8008
  stx $4300		;Set DMA mode to fixed source, BYTE to $2180
  ldx #.LOWORD(zero_fill_byte)
  stx $4302		;Set source offset
  lda #.BANKBYTE(zero_fill_byte)
  sta $4304		;Set source bank
  ldx #$0000
  stx $4305		;Set transfer size to 64KBytes (65536 bytes)
  lda #$01
  sta $420B		;Initiate transfer

  lda #$01		;now zero the next 64KB (i.e. 128KB total)
  sta $420B		;Initiate transfer

  rep #$30	;A=16, X/Y=16
  sep #$20	;A=8

  jsr initialize_oam
  jsr initialize_controllers

  sep #$20	;A=8

  lda #$F0		;Increment VRAM address by 1 after write to $2119
  sta $2115
  ldx #$0000
  stx $2116		;$2116: Word address for accessing VRAM

  ldx #.LOWORD(BallTile)
  stx $4302		;Store Data offset into DMA source offset
  lda #.BANKBYTE(BallTile)
  sta $4304		;Store data Bank into DMA source bank
  ldx #$20
  stx $4305		;Store size of data block

  lda #$01
  sta $4300	;Set DMA mode (word, normal increment)
  lda #$18	;Set the destination register (VRAM write register)
  sta $4301
  lda #$01	;Initiate DMA transfer (channel 1)
  sta $420B

  lda #$F0		;Increment VRAM address by 1 after write to $2119
  sta $2115
  ldx #$0010
  stx $2116		;$2116: Word address for accessing VRAM

  ldx #.LOWORD(PaddleTile)
  stx $4302		;Store Data offset into DMA source offset
  lda #.BANKBYTE(PaddleTile)
  sta $4304		;Store data Bank into DMA source bank
  ldx #$20
  stx $4305		;Store size of data block

  lda #$01
  sta $4300	;Set DMA mode (word, normal increment)
  lda #$18	;Set the destination register (VRAM write register)
  sta $4301
  lda #$01	;Initiate DMA transfer (channel 1)
  sta $420B

  jsr setup_video	;setup the video modes and such, then turn on the screen

;Enable NMI (when interrupts get re-enabled) and joypad reading
  sep #$20	;A=8
  lda #$81
  sta $4200

  cli			;Restore interrupts -- we're all ready to go!

  lda #$81
  sta $2121
  lda #$FF
  sta $2122
  lda #$FF
  sta $2122

  rep #$30	;A=16, X/Y=16

  lda #40
  sta BallXPosition
  lda #108
  sta BallYPosition

  lda #96
  sta Paddle1YPosition
  lda #96
  sta Paddle2YPosition

  lda #$0020
  sta SpriteBuf1+4
  lda #$0001
  sta SpriteBuf1+6
  stz SpriteBuf3+5
  lda #$0020
  sta SpriteBuf1+8
  lda #$0001
  sta SpriteBuf1+10
  stz SpriteBuf3+9
  lda #$0020
  sta SpriteBuf1+12
  lda #$0001
  sta SpriteBuf1+14
  stz SpriteBuf3+13
  lda #$0020
  sta SpriteBuf1+16
  lda #$0001
  sta SpriteBuf1+18
  stz SpriteBuf3+17

  lda #$00D8
  sta SpriteBuf1+20
  lda #$0001
  sta SpriteBuf1+22
  stz SpriteBuf3+21
  lda #$00D8
  sta SpriteBuf1+24
  lda #$0001
  sta SpriteBuf1+26
  stz SpriteBuf3+25
  lda #$00D8
  sta SpriteBuf1+28
  lda #$0001
  sta SpriteBuf1+30
  stz SpriteBuf3+29
  lda #$00D8
  sta SpriteBuf1+32
  lda #$0001
  sta SpriteBuf1+34
  stz SpriteBuf3+33

  lda #$0002
  sta BallXVelocity
  sta BallYVelocity

  lda #%0000000000000010
  sta GameState

;=========================================================================================

infinite_loop:
  wai			;Wait for interrupts to finish (NMI/VBlank)

  rep #$30	;A=16, X/Y=16

  lda Joy1Data
  and #WordUpDPad
  beq done_checking_paddle_1_up
  lda Paddle1YPosition
  beq done_checking_paddle_1_up
  sec
  sbc #$0003
  sta Paddle1YPosition
done_checking_paddle_1_up:

  lda Joy1Data
  and #WordDownDPad
  beq done_checking_paddle_1_down
  lda Paddle1YPosition
  cmp #192
  beq done_checking_paddle_1_down
  clc
  adc #$0003
  sta Paddle1YPosition
done_checking_paddle_1_down:

  lda Joy2Data
  and #WordUpDPad
  beq done_checking_paddle_2_up
  lda Paddle2YPosition
  beq done_checking_paddle_2_up
  sec
  sbc #$0003
  sta Paddle2YPosition
done_checking_paddle_2_up:

  lda Joy2Data
  and #WordDownDPad
  beq done_checking_paddle_2_down
  lda Paddle2YPosition
  cmp #192
  beq done_checking_paddle_2_down
  clc
  adc #$0003
  sta Paddle2YPosition
done_checking_paddle_2_down:

;=========================================================================================

  lda GameState
  and #%0000000000000010
  beq p1_not_with_ball

  lda Paddle1YPosition
  clc
  adc #12
  sta BallYPosition
  
  lda Joy1Data
  and #AButton
  bne p1_with_ball
  brl done_moving_ball

p1_with_ball:
  lda GameState
  and #%1111111111111101
  sta GameState
  bra p2_not_with_ball

p1_not_with_ball:

  lda GameState
  and #%0000000000000100
  beq p2_not_with_ball

  lda Paddle2YPosition
  clc
  adc #12
  sta BallYPosition

  lda Joy2Data
  and #AButton
  bne p2_with_ball
  brl done_moving_ball

p2_with_ball:
  lda GameState
  and #%1111111111111011
  sta GameState

p2_not_with_ball:

;=========================================================================================

  lda BallXPosition
  bne right_wall_collision_check
  lda GameState
  ora #%0000000000000100
  sta GameState
  lda #208
  sta BallXPosition
  lda Paddle2YPosition
  clc
  adc #12
  sta BallYPosition
  bra done_moving_ball

right_wall_collision_check:
  cmp #248
  bcc continue_finding_ball_x_position
  lda GameState
  ora #%0000000000000010
  sta GameState
  lda #40
  sta BallXPosition
  lda Paddle1YPosition
  clc
  adc #12
  sta BallYPosition
  bra done_moving_ball
  
continue_finding_ball_x_position:
  lda BallXPosition
  clc
  adc BallXVelocity
  sta BallXPosition

  lda BallYPosition
  cmp #216
  bcc continue_finding_ball_y_position
  lda BallYVelocity
  eor #$FFFF
  inc
  sta BallYVelocity

continue_finding_ball_y_position:
  lda BallYPosition
  clc
  adc BallYVelocity
  sta BallYPosition

done_moving_ball:

;=========================================================================================

  lda BallXPosition
  cmp #$0028
  bne done_checking_paddle_1_collision
  lda BallYPosition
  cmp Paddle1YPosition
  bcc done_checking_paddle_1_collision
  lda BallYPosition
  sec
  sbc #$0020
  cmp Paddle1YPosition
  bcs done_checking_paddle_1_collision
  lda BallXVelocity
  eor #$FFFF
  inc
  sta BallXVelocity

done_checking_paddle_1_collision:

  lda BallXPosition
  cmp #$00D8
  bne done_checking_paddle_2_collision
  lda BallYPosition
  cmp Paddle2YPosition
  bcc done_checking_paddle_2_collision
  lda BallYPosition
  sec
  sbc #$0020
  cmp Paddle2YPosition
  bcs done_checking_paddle_2_collision
  lda BallXVelocity
  eor #$FFFF
  inc
  sta BallXVelocity

done_checking_paddle_2_collision:

;=========================================================================================

  sep #$20	;A=8
  lda BallXPosition
  sta SpriteBuf1
  lda BallXPosition+1
  and #$01
  sta SpriteBuf3+1
  lda BallYPosition
  sta SpriteBuf1+1

  lda Paddle1YPosition
  sta SpriteBuf1+5
  lda Paddle1YPosition
  clc
  adc #$08
  sta SpriteBuf1+9
  lda Paddle1YPosition
  clc
  adc #$10
  sta SpriteBuf1+13
  lda Paddle1YPosition
  clc
  adc #$18
  sta SpriteBuf1+17

  lda Paddle2YPosition
  sta SpriteBuf1+21
  lda Paddle2YPosition
  clc
  adc #$08
  sta SpriteBuf1+25
  lda Paddle2YPosition
  clc
  adc #$10
  sta SpriteBuf1+29
  lda Paddle2YPosition
  clc
  adc #$18
  sta SpriteBuf1+33

  lda #$0024
  sta SpriteCount

  jsr process_oam

  jmp infinite_loop	;Do this forever
.endproc

;=========================================================================================

.proc VBlank
  rep #$30	;A=16, X/Y=16
  pha
  phx
  phy
  phd
  phb
  php
  lda #$0000
  tcd

;transfer sprite data

  sep #$20	;A=8
  stz $2102	;set OAM address to 0
  stz $2103	;...and no priority rotation

  stz $4300			;CPU -> PPU, auto-increment, write once
  lda #$04
  sta $4301			;DMA will write to register $2104 (OAM data write)
  ldy #.LOWORD(SpriteBuf1)
  sty $4302			;source offset of SpriteBuf1
  lda #.BANKBYTE(SpriteBuf1)
  sta $4304			;bank address of SpriteBuf1
  ldy #SpriteBufSize
  sty $4305			;transfer SpriteBufSize bytes (pre-calculated number)
  lda #$01
  sta $420B		;start DMA transfer

  jsr get_input	;update joypad data

  lda $4210	;clear NMI Flag

  plp
  plb
  pld
  ply
  plx
  pla
  rti
.endproc

;=========================================================================================

.proc clear_vram
  rep #$30	;A=16, X/Y=16
  sep #$20	;A=8
  lda #$80
  sta $2115		;Set VRAM port to word access
  ldx #$1809
  stx $4300		;Set DMA mode to fixed source, WORD to $2118/9
  ldx #$0000
  stx $2116			;Set VRAM port address to $0000
  ldx #.LOWORD(zero_fill_byte)
  stx $4302			;Set source address to $xx:0000
  lda #.BANKBYTE(zero_fill_byte)
  sta $4304			;Set source bank
  ldx #$0000
  stx $4305		;Set transfer size to 65536 bytes
  lda #$01
  sta $420B		;Initiate transfer
  rts
.endproc

;=========================================================================================

.proc clear_cgram
  rep #$30	;A=16, X/Y=16
  sep #$20	;A=8
  stz $2121
  ldx #$0100

clear_palette_loop:
  stz $2122
  stz $2122
  dex
  bne clear_palette_loop
  rts
.endproc

;=========================================================================================

.proc initialize_oam
  php	
  rep #$30

  lda #$01
  ldx #0

initialize_oam_loop:
  sta SpriteBuf3+1,x
  inx
  inx
  inx
  inx
  cpx #$0200
  bne initialize_oam_loop
  plp
  rts
.endproc

;=========================================================================================

.proc initialize_controllers
  php	

  rep	#$10	;8 bit mem/A, 16 bit X/Y
  sep	#$20

  lda #$C0	;have the automatic read of the SNES read the first pair of JoyPads
  sta WRIO

  ldx #$0000
  stx Joy1Press
  stx Joy2Press

;in ZSNES, the auto-read doesn't work unless NMI is enabled... is this correct?
;
; lda $4212
; lda $4212
;
;w11:
; lda $4212
; and #$01
; beq _w11	;Wait till automatic JoyPort read starts
;
; jsr GetInput	;fill the variables with real Joypad data
;
;w12:
; lda $4212
; and #$01
; beq _w12	;Wait till automatic JoyPort read starts
;
; jsr get_input	;fill the variables with real Joypad data

  plp
  rts
.endproc

;=========================================================================================

.proc get_input
  php
  sep #$20	;A=8

  lda #$01
: bit $4212
  bne :-		;Wait till automatic JoyPort read is complete

  rep #$30   ;A=16, X/Y=16

;Get Joypads 1,2

  lda JOY0		;get JoyPad1
  tax 			
  eor Joy1Data		;A = A xor JoyState = (changes in joy state)
  stx Joy1Data		;update JoyState
  ora Joy1Press		;A = (joy changes) or (buttons pressed)
  and Joy1Data		;A = ((joy changes) or (buttons pressed)) and (current joy state)  
  sta Joy1Press		;store A = (buttons pressed since last clearing reg) and (button is still down)

  lda JOY1		;get JoyPad2
  tax 			
  eor Joy2Data		;A = A xor JoyState = (changes in joy state)
  stx Joy2Data		;update JoyState
  ora Joy2Press		;A = (joy changes) or (buttons pressed)
  and Joy2Data		;A = ((joy changes) or (buttons pressed)) and (current joy state)  
  sta Joy2Press		;store A = (buttons pressed since last clearing reg) and (button is still down)

;Make sure Joypads 1,2 are valid

  sep #$30	;A=8, X/Y=8
  lda JOYSER0
  eor #$01
  and #$01		;A = -bit0 of JoySer0
  ora Joy1Data
  sta Joy1Data		;joy state = (joy state) or A.... so bit0 of Joy1State = 0 only if it is a valid joypad

  lda JOYSER1
  eor #$01
  and #$01		;A = -bit0 of JoySer1
  ora Joy2Data		
  sta Joy2Data		;joy state = (joy state) or A.... so bit0 of Joy1State = 0 only if it is a valid joypad

;Change all invalid joypads to have a state of no button presses

  rep #$30	;A=16, X/Y=16

  ldx #$0001
  lda #$000f

  bit Joy1Data		;A = joy state, if any of the bottom 4 bits are on... either nothing is plugged
  beq :+		;into the joy port, or it is not a joypad
  stx Joy1Data		;if it is not a valid joypad put $0001 into the 2 joy state variables
  stz Joy1Press

: bit Joy2Data		;A = joy state, if any of the bottom 4 bits are on... either nothing is plugged
  beq done		;into the joy port, or it is not a joypad
  stx Joy2Data		;if it is not a valid joypad put $0001 into the 2 joy state variables
  stz Joy2Press

done:
  plp
  rts
.endproc

;=========================================================================================

.proc setup_video
  php
  rep #$30	;A=16, X/Y=16
  sep #$20	;A=8

  stz $211b
  lda #$01
  sta $211b
  stz $211c
  stz $211c
  stz $211d
  stz $211d
  stz $211e
  sta $211e
  lda #$80
  sta $211f
  stz $211f
  sta $2120
  stz $2120

  lda #%00000000	;Sprites 8x8 or 16x16, character data at $0000 (word address)
  sta $2101         

  lda #$01		;Set video mode 1, 16x16 tiles (16 color BG1, 16 color BG2, 4 color BG3)
  sta $2105         

  lda #$70		;Set BG1's Tile Map VRAM offset to $7800 (word address)
  sta $2107		;and the Tile Map size to 32 tiles x 32 tiles

  lda #$02		;Set BG1's Character VRAM offset to $4000 (word address)
  sta $210B

  lda #$10		;Turn on Sprites
  sta $212C

  lda #$0f		;Turn on screen, full brightness
  sta $2100

  plp
  rts
.endproc

;=========================================================================================

.proc process_oam
  php
  rep #$10
  sep #$20

  ldx SpriteCount
  lda #$01

clear_oam_loop:
  cpx #$0200
  beq finished_clearing_oam_loop
  sta SpriteBuf1,x
  stz SpriteBuf1+1,x
  stz SpriteBuf1+2,x
  stz SpriteBuf1+3,x
  sta SpriteBuf3+1,x		;sprite 3, 9th x bit
  inx
  inx
  inx
  inx
  bra clear_oam_loop

finished_clearing_oam_loop:
  ldy #$0000
  sty SpriteCount

build_hioam_loop:
  lda SpriteBuf3+14,y		;sprite 3, size bit
  asl
  ora SpriteBuf3+13,y		;sprite 3, 9th x bit
  asl
  ora SpriteBuf3+10,y		;sprite 2, size bit
  asl
  ora SpriteBuf3+9,y		;sprite 2, 9th x bit
  asl
  ora SpriteBuf3+6,y		;sprite 1, size bit
  asl
  ora SpriteBuf3+5,y		;sprite 1, 9th x bit
  asl
  ora SpriteBuf3+2,y		;sprite 0, size bit
  asl
  ora SpriteBuf3+1,y		;sprite 0, 9th x bit
  sta SpriteBuf1,x
  rep #$20
  tya
  clc
  adc #$0010
  tay
  sep #$20
  inx
  cpx #$0220
  bne build_hioam_loop

  plp
  rts
.endproc

;=========================================================================================
.segment "RODATA"
;=========================================================================================

BallTile:
  .incbin "Ball.pic"

PaddleTile:
  .incbin "Paddle.pic"

;Byte used to zero/initialise all of memory.
;This is needed because DMA requires an address of a byte to
;read from, not a raw value.

zero_fill_byte:
  .byte $00